# date4/timezone/tzfile.rb: Written by Tadayoshi Funaba 2006
# $Id: tzfile.rb,v 1.3 2006-12-30 21:50:28+09 tadf Exp $

require 'date4'

class Date4

  module TimeZone

    class TZFile

      class ParseError < StandardError; end

      def self.zoneinfodirs
	%w(/usr/share
	   /usr/share/lib
	   /usr/lib
	   /usr/local/share
	   /usr/local/share/lib
	   /usr/local/lib).
	  collect{|pre| pre + '/zoneinfo'}
      end

      def self.zoneinfodir
	@@zoneinfodir ||= zoneinfodirs.detect{|dir| File.directory? dir}
      end

      def initialize(path)
	@path = path
	begin
	  fo = open(File.expand_path(@path, self.class.zoneinfodir))

	  unless fo.read(4) == 'TZif'
	    raise ParseError, 'not a tzfile'
	  end

	  oldver = fo.read(1) == "\000"

	  fo.read(15)

	  (if oldver then 1 else 2 end).times do |nth|

	    @ttisgmtcnt, @ttisstdcnt, @leapcnt, @timecnt, @typecnt, @charcnt =
	      fo.read(6 * 4).unpack('N*')

	    unless @leapcnt == 0
	      raise ParseError, 'does not support leap seconds'
	    end

	    @trans = (0...@timecnt).collect do
	      if nth == 0
		x, = fo.read(4).unpack('N')
		x -= 2**32 if x >= 2**31
	      else
		a, b, = fo.read(8).unpack('NN')
		x = a << 32 | b
		x -= 2**64 if x >= 2**63
	      end
	      x
	    end
	    @index = (0...@timecnt).collect{fo.read(1).unpack('C')[0]}
	    @transp = @trans.zip(@index)

	    @local = (0...@typecnt).collect do
	      xs = fo.read(6).unpack('NC2')
	      xs[0] -= 2**32 if xs[0] >= 2**31
	      xs
	    end

	    @abbr  = fo.read(@charcnt)

	    @leap  = (0...@leapcnt).collect do
	      if nth == 0
		x = fo.read(4).unpack('N')
		x -= 2**32 if x >= 2**31
	      else
		a, b = fo.read(8).unpack('N2')
		x = a << 32 | b
		x -= 2**64 if x >= 2**63
	      end
	      y = fo.read(4).unpack('N')
	      y -= 2**32 if y >= 2**31
	      [x, y]
	    end

	    @isstd = (0...@ttisstdcnt).collect{fo.read(1).unpack('C')[0]}
	    @isgmt = (0...@ttisgmtcnt).collect{fo.read(1).unpack('C')[0]}

	    if nth == 0
	      unless oldver
		fo.read(20) # skip the second header
	      end
	    end

	  end

	  unless oldver
	    @posix = fo.read.strip
	  end

	ensure
	  fo.close if fo
	end
      end

      def search_segment(a)
	l = 0
	u = a.size
	while l < u
	  m = ((l + u) / 2).truncate
	  if (yield a[m]) <= 0
	    l = m + 1
	  else
	    u = m
	  end
	end
	l - 1
      end

      def lookup(ajd, adj)
	j = search_segment(@transp) do |t, i|
	  if adj
	    of = (@local[i][0]).to_r/86400
	  else
	    of = 0
	  end
	  ((4881175.to_r/2) + t.to_r/86400) <=> (ajd - of)
	end

	if j >= 0
	  i2 = @transp[j][1]
	else
	  i2 = 0
	  @local.size.times do |i|
	    if @local[i][1] == 0
	      i2 = i
	      break
	    end
	  end
	end

	local = @local[i2]
	zone, = @abbr[local[2]..-1].split("\000")
	of = local[0].to_r/86400
	utc = false
	dst = local[1] == 1

	{:zone=>zone, :of=>of, :utc=>utc, :dst=>dst}
      end

      def inspect() format('#<%s: %s>', self.class, @path) end
      def to_s() @path.dup end

=begin
      def _dump(limit) Marshal.dump(@path, -1) end
      def self._load(str) new(Marshal.load(str)) end
=end

    def marshal_dump() @path end
    def marshal_load(path) initialize(path) end

    end

  end

end
